# -*- coding: utf-8 -*-
"""
Mutualistic Multi-Optimization
"""

import numpy as np
from .optimizer import Optimizer, CandidateState
from scipy.interpolate import interp1d # need this for akb_model
from scipy.stats import cauchy


class Particle(CandidateState):
    """PSO Particle class"""
    
    def __init__(self, optimizer: Optimizer):
        CandidateState.__init__(self, optimizer)        
        self.V = np.zeros([optimizer.dimensions]) * np.nan
        
        
class Solution(CandidateState):
    """DE solution class"""
    
    def __init__(self, optimizer: Optimizer):
        CandidateState.__init__(self, optimizer)
        #super(Particle, self).__init__(optimizer) # ugly version of the above
        
        self.CR = None
        self.F = None
        self.V = np.zeros([optimizer.dimensions]) * np.nan # mutant vector


class MMO(Optimizer):
    """Mutualistic Multi-Optimization class"""

    def __init__(self):
        """Initialization"""
        Optimizer.__init__(self)

        self.X0 = None
        self.params = {}        
        self.method = {}
        
        # PSO
        # no special parameters
        
        # FWA  
        self.cX = None


    def _check_params(self):
        defined_params = list(self.params.keys())
        mandatory_params, optional_params = [], []
        
        if not self.method:
            self.method = {'PSO': 'Vanilla', 
                           'FWA': 'Vanilla'}
        assert len(self.method) >= 2, \
            'optimizer.method should provide at least 2 optimization methods'
        
        if 'PSO' in self.method:        
    
            if self.method['PSO'] == 'Vanilla':
                mandatory_params += 'swarm_size inertia cognitive_rate social_rate'.split()
                if 'swarm_size' not in self.params:
                    self.params['swarm_size'] = self.dimensions
                    defined_params += 'swarm_size'.split()
                if 'inertia' not in self.params:
                    self.params['inertia'] = 0.72
                    defined_params += 'inertia'.split()
                if 'cognitive_rate' not in self.params:
                    self.params['cognitive_rate'] = 1.0
                    defined_params += 'cognitive_rate'.split()
                if 'social_rate' not in self.params:
                    self.params['social_rate'] = 1.0
                    defined_params += 'social_rate'.split()
                optional_params += 'akb_model akb_fun_start akb_fun_stop'.split()
            elif self.method['PSO'] == 'TVAC':
                mandatory_params += 'swarm_size inertia'.split()
                if 'swarm_size' not in self.params:
                    self.params['swarm_size'] = self.dimensions
                    defined_params += 'swarm_size'.split()
                if 'inertia' not in self.params:
                    self.params['inertia'] = 0.72
                    defined_params += 'inertia'.split()
                optional_params += 'akb_model akb_fun_start akb_fun_stop'.split()
            else:
                assert False, f"Unknown PSO method! {self.method['PSO']}"
        
        if 'FWA' in self.method:
            
            if self.method['FWA'] == 'Vanilla':
                mandatory_params += 'n m1 m2'.split()
                if 'n' not in self.params:
                    self.params['n'] = self.dimensions
                    defined_params += 'n'.split()
                if 'm1' not in self.params:
                    self.params['m1'] = self.dimensions // 2
                    defined_params += 'm1'.split()
                if 'm2' not in self.params:
                    self.params['m2'] = self.dimensions // 2
                    defined_params += 'm2'.split()    
                optional_params += ''.split()
            elif self.method['FWA'] == 'Rank':
                mandatory_params += 'n m1 m2'.split()
                if 'n' not in self.params:
                    self.params['n'] = self.dimensions
                    defined_params += 'n'.split()
                if 'm1' not in self.params:
                    self.params['m1'] = self.dimensions // 2
                    defined_params += 'm1'.split()
                if 'm2' not in self.params:
                    self.params['m2'] = self.dimensions // 2
                    defined_params += 'm2'.split()
                optional_params += ''.split()
            else:
                assert False, f"Unknown FWA method! {self.method['FWA']}"
                
        if 'DE' in self.method:
            
            if self.method['DE'] == 'SHADE':
                mandatory_params += 'initial_population_size external_archive_size_factor historical_memory_size p_mutation'.split()
                if 'initial_population_size' not in self.params:
                    self.params['initial_population_size'] = self.dimensions * 18
                    defined_params += 'initial_population_size'.split()
                if 'external_archive_size_factor' not in self.params:
                    self.params['external_archive_size_factor'] = 2.6
                    defined_params += 'external_archive_size_factor'.split()
                if 'historical_memory_size' not in self.params: # a.k.a. H
                    self.params['historical_memory_size'] = 6
                    defined_params += 'historical_memory_size'.split()
                if 'p_mutation' not in self.params:
                    self.params['p_mutation'] = 0.11
                    defined_params += 'p_mutation'.split()    
                optional_params = ''.split()
            elif self.method['DE'] == 'LSHADE':
                mandatory_params += 'initial_population_size external_archive_size_factor historical_memory_size p_mutation'.split()
                if 'initial_population_size' not in self.params:
                    self.params['initial_population_size'] = self.dimensions * 18
                    defined_params += 'initial_population_size'.split()
                if 'external_archive_size_factor' not in self.params:
                    self.params['external_archive_size_factor'] = 2.6
                    defined_params += 'external_archive_size_factor'.split()
                if 'historical_memory_size' not in self.params: # a.k.a. H
                    self.params['historical_memory_size'] = 6
                    defined_params += 'historical_memory_size'.split()
                if 'p_mutation' not in self.params:
                    self.params['p_mutation'] = 0.11
                    defined_params += 'p_mutation'.split()  
                optional_params = ''.split()
            else:
                assert False, f"Unknown DE method! {self.method['DE']}"

        if 'PSO' in self.method:
            
            """ Anakatabatic Inertia a.k.a. Polynomial PFIDI """
            if self.params['inertia'] == 'anakatabatic':
                assert ('akb_fun_start' in defined_params \
                        and 'akb_fun_stop' in defined_params) \
                        or 'akb_model' in defined_params, \
                        'Error: Anakatabatic inertia requires either akb_model parameter or akb_fun_start and akb_fun_stop parameters'
                optional_params += 'akb_fun_start akb_fun_stop'.split()
                
                if 'akb_model' in defined_params:                    
                    optional_params += 'akb_model'.split()
    
                    if self.params['akb_model'] in ['FlyingStork', 'MessyTie', 'RightwardPeaks', 'OrigamiSnake']:   # w-list-based named akb_models                
                        if self.params['akb_model'] == 'FlyingStork':
                            w_start = [-0.86, 0.24, -1.10, 0.75, 0.72]
                            w_stop = [-0.81, -0.35, -0.26, 0.64, 0.60]
                            splinetype = 'linear'
                            if self.method['PSO'] != 'Vanilla':
                                print('Warning: akb_model \'FlyingStork\' was designed for Vanilla PSO')                    
                        elif self.params['akb_model'] == 'MessyTie':
                            w_start = [-0.62, 0.18, 0.65, 0.32, 0.77]
                            w_stop = [0.36, 0.73, -0.62, 0.40, 1.09]
                            splinetype = 'linear'
                            if self.method['PSO'] != 'Vanilla':
                                print('Warning: akb_model \'MessyTie\' was designed for Vanilla PSO')   
                        elif self.params['akb_model'] == 'RightwardPeaks':
                            w_start = [-1.79, -0.33, 2.00, -0.67, 1.30]
                            w_stop = [-0.91, -0.88, -0.84, 0.67, -0.36]
                            splinetype = 'linear'
                            if self.method['PSO'] != 'TVAC':
                                print('Warning: akb_model \'RightwardPeaks\' was designed for TVAC PSO')
                        elif self.params['akb_model'] == 'OrigamiSnake':
                            w_start = [-1.36, 2.00, 1.00, -0.60, 1.22]
                            w_stop = [0.30, 1.03, -0.21, 0.40, 0.06]
                            splinetype = 'linear'
                            if self.method['PSO'] != 'TVAC':
                                print('Warning: akb_model \'OrigamiSnake\' was designed for TVAC PSO')                      # code shared for all w-list-based named akb_models
                        Th = np.linspace(np.pi/4, 5*np.pi/4, 5)
                        self.params['akb_fun_start'] = \
                                            interp1d(Th, w_start, kind=splinetype)
                        self.params['akb_fun_stop'] = \
                                            interp1d(Th, w_stop, kind=splinetype) 
                    else:
                        if self.params['akb_model'] != 'Languid':
                            print('Warning: Unknown akb_model. Defaulting to \'Languid\'')
                            self.params['akb_model'] = 'Languid'
                    if self.params['akb_model'] == 'Languid':
                        def akb_fun_languid(Th):
                            w = (0.72 + 0.05) * np.ones_like(Th)
                            for i, th in enumerate(Th):
                                if th < 4*np.pi/4: 
                                    w[i] = 0
                            return w
                        self.params['akb_fun_start'] = akb_fun_languid
                        self.params['akb_fun_stop'] = akb_fun_languid 
        
        Optimizer._check_params(self, mandatory_params, optional_params, defined_params)

        
    def _init_method(self):
        self.lb = np.array(self.lb)
        self.ub = np.array(self.ub)
               
        if 'PSO' in self.method:
            
            """ Initialize PSO """
            # Bounds for velocity
            self.v_max = 0.2 * (self.ub - self.lb)
    
            # Generate a swarm
            self.cS = np.array([Particle(self) for c in range(self.params['swarm_size'])], dtype=Particle)
            
            # Prepare arrays
            self.dF = np.empty([self.params['swarm_size']]) * np.nan
    
            # Generate initial positions
            for p in range(self.params['swarm_size']):
                
                # Random position
                self.cS[p].X = np.random.uniform(self.lb, self.ub)
                
                # Using specified particles initial positions
                if self.X0:
                    if p < np.shape(self.X0)[0]:
                        self.cS[p].X = self.X0[p]
                        
                # Generate velocity
                self.cS[p].V = np.random.uniform(-self.v_max, self.v_max)
    
                # No fitness change at the start
                self.dF[p] = 0.0
    
            # Evaluate
            self.collective_evaluation(self.cS)
            
            # Use initial particles as best ones
            self.cB = np.array([cP.copy() for cP in self.cS], dtype=CandidateState)
            
            # Update the overall best
            self.p_best = np.argmin(self.cB)
                
            # # Update history
            # #self.results.cHistory = [self.cB[self.p_best].copy()] # superseded by progress_log()
            # self.progress_log()
            
            self.BI = np.zeros(self.params['swarm_size'], dtype=int)
            self.TOPO = np.zeros([self.params['swarm_size'], self.params['swarm_size']], dtype=np.bool)
    
            self.reinitialize_topology()
            self.find_neighbourhood_best()
        
        if 'FWA' in self.method:
            
            """ Initialize FWA """
            self.cX = np.array([CandidateState(self) for p in range(self.params['n'])], dtype=CandidateState)
            
            # Generate initial positions
            for p in range(self.params['n']):
                
                # Random position
                self.cX[p].X = np.random.uniform(self.lb, self.ub, self.dimensions)
                
                # Using specified initial positions
                if self.X0:
                    if p < np.shape(self.X0)[0]:
                        self.cX[p].X = self.X0[p]
    
            # Evaluate all
            self.collective_evaluation(self.cX)
    
            # Sort
            self.cX = np.sort(self.cX)
        
        if 'DE' in self.method:
            
            """ Initialize DE """
            # Generate a population
            self.Pop = np.array([Solution(self) for c in \
                                 range(self.params['initial_population_size'])], dtype=Solution)
            
            # Generate a trial population
            self.Trials = np.array([Solution(self) for c in \
                                    range(self.params['initial_population_size'])], dtype=Solution)
            
            # Initalize Archive
            self.A = np.empty([0])
            
            # Prepare historical memory
            self.M_CR = np.full(self.params['historical_memory_size'], 0.5)
            self.M_F = np.full(self.params['historical_memory_size'], 0.5)
    
            # Generate initial positions
            for i in range(self.params['initial_population_size']):
                
                # Random position
                self.Pop[i].X = np.random.uniform(self.lb, self.ub)
                
                # Using specified particles initial positions
                if self.X0:
                    if i < np.shape(self.X0)[0]:
                        self.Pop[i].X = self.X0[i]

            # Evaluate
            self.collective_evaluation(self.Pop)
        
        """ After all methods are initialized... """
        # Update history
        self._progress_log()
 
       
    """ PSO functions """
        
    def reinitialize_topology(self, k=3):
        self.TOPO[:, :] = False
        for p in range(self.params['swarm_size']):
            links = np.random.randint(self.params['swarm_size'], size=k)
            self.TOPO[p, links] = True
            self.TOPO[p, p] = True  

    def find_neighbourhood_best(self):
        for p in range(self.params['swarm_size']):
            links = np.where(self.TOPO[p, :])[0]
            #best = np.argmin(self.BF[links])
            p_best = np.argmin(self.cB[links])
            p_best = links[p_best]
            self.BI[p] = p_best


    """ FWA functions """
        
    def explosion(self):
        eps=0.001
        amp=10
        a=0.01
        b=10
        F = np.array([cP.f for cP in self.cX])
        fmin = np.min(F)
        fmax = np.max(F)
        
        explosion_sparks = []
        for p in range(self.params['n']):
               
            cFw = self.cX[p].copy()
            #print(cFw.X)
            
            if self.method['FWA'] == 'Vanilla':
                # Number of sparks
                n1 = self.params['m1'] * (fmax - cFw.f + eps) / np.sum(fmax - F + eps)
                n1 = self.min_max_round(n1, self.params['m1'] * a, self.params['m2'] * b)
                
                # Amplitude
                A = amp * (cFw.f - fmin + eps) /  (np.sum(F - fmin) + eps)

                for j in range(n1):
                    for k in range(self.dimensions):
                        if (np.random.choice([True, False])):
                            cFw.X[k] += np.random.uniform(-A, A)
                    explosion_sparks.append(cFw.copy())
                
            if self.method['FWA'] == 'Rank':
                
                # Number of sparks
                #vn1 = self.params['m1'] * (fmax - cFw.f + eps) / np.sum(fmax - F + eps)
                #vn1 = self.min_max_round(vn1, self.params['m1'] * a, self.params['m2'] * b)
                
                n1 = self.params['m1'] * (self.params['n'] - p)**1 / np.sum(np.arange(self.params['n']+1)**1)
                n1 = np.random.choice([int(np.floor(n1)), int(np.ceil(n1))])
                #print(self.cX[p].f, vn1, n1)
                
                # Amplitude
                #Ac = amp * (cFw.f - fmin + eps) / (np.sum(F - fmin) + eps)
                    
                #print('n1:', n1, 'A:', A)
                XX = np.array([cP.X for cP in self.cX])
                #print(XX.shape)
                
                # Uniform
                dev = np.std(XX, 0)
                avg_scale = np.average(np.sqrt(np.arange(self.params['n']) + 1))
                scale = np.sqrt(p + 1) / avg_scale
                
                #avg_scale = np.average(np.arange(self.params['n']) + 1)
                #scale = (p + 1) / avg_scale
                                
                A = np.sqrt(12) / 2 * dev * scale
                A *= 1.5
              
                #cS = cFw.copy()
                for j in range(n1):
                    cFw.X = cFw.X + np.random.uniform(-A, A) * np.random.randint(0, 1, A.size)
                    
                    for k in range(self.dimensions):
                        if (np.random.choice([True, False])):
                            # Uniform
                            cFw.X[k] += np.random.uniform(-A[k], A[k])
                            # Normal
                            # cFw.X[k] += np.random.normal(-A[k], A[k])
                    
                    #print(cS.X)
                    explosion_sparks.append(cFw.copy())  

        return explosion_sparks        
    
    def gaussian_mutation(self):
        mutation_sparks = []
        for j in range(self.params['m2']):
            cFw = self.cX[np.random.randint(self.params['n'])].copy()
            g = np.random.normal(1, 1)
            for k in range(self.dimensions):
                if(np.random.choice([True, False])):
                    cFw.X[k] *= g
            mutation_sparks.append(cFw)
        return mutation_sparks  

    def __mapping_rule(self, sparks, lb, ub, dimension):
        for i in range(len(sparks)):
            for j in range(dimension):
                if(sparks[i].X[j] > ub[j] or sparks[i].X[j] < lb[j]):
                    sparks[i].X[j] = lb[j] + \
                        (sparks[i].X[j] - lb[j]) % (ub[j] - lb[j])
  
    def __selection(self, sparks, n, function):        
        self.collective_evaluation(sparks)
        self.cX = np.append(self.cX, sparks)   

    def min_max_round(self, s, smin, smax):
        return int(np.round(np.min([np.max([s, smin]), smax])))   

    def __round(self, s, m, a, b):
        if (s < a * m):
            return round(a * m)
        elif (s > b * m):
            return round(b * m)
        else:
            return round(s)
    
    """ DE functions """
    
    # no special DE functions
    

    """ MMO run function """
    def _run(self):
        self._check_params()
        self._init_method()

        """ Prepare PSO """
        if 'PSO' in self.method:   
            if 'inertia' in self.params.keys():
                w = self.params['inertia']
            if 'cognitive_rate' in self.params.keys():
                c1 = self.params['cognitive_rate']
            if 'cognitive_rate' in self.params.keys():
                c2 = self.params['social_rate']

        """ Prepare FWA """
        if 'FWA' in self.method:
            n = self.params['n']
        
        """ Prepare DE """
        # no special preparement needed
        
        """ Main run loop """
        for self.it in range(1, self.iterations + 1):
            
            """ PSO iteration """
            if 'PSO' in self.method:
                R1 = np.random.uniform(0, 1, [self.params['swarm_size'], self.dimensions])
                R2 = np.random.uniform(0, 1, [self.params['swarm_size'], self.dimensions])
    
                if self.params['inertia'] == 'LDIW':
                    w = 1.0 - (1.0 - 0.4) * self.it / self.iterations
    
                if self.method['PSO'] == 'TVAC':
                    c1 = 2.5 - (2.5 - 0.5) * self.it / self.iterations
                    c2 = 0.5 + (2.5 - 0.5) * self.it / self.iterations
    
                if self.params['inertia'] == 'anakatabatic':
    
                    theta = np.arctan2(self.dF, np.min(self.dF))
                    theta[theta < 0] = theta[theta < 0] + 2 * np.pi  # 3rd quadrant
                    # fix for atan2(0,0)=0
                    theta0 = theta < 1e-300
                    theta[theta0] = np.pi / 4 + \
                        np.random.rand(np.sum(theta0)) * np.pi
                    w_start = self.params['akb_fun_start'](theta)
                    w_stop = self.params['akb_fun_stop'](theta)
                    #print(w_start)
                    w = w_start * (1 - self.it / self.iterations) \
                        + w_stop * (self.it / self.iterations)
                
                w = w * np.ones(self.params['swarm_size']) # ensure w is a vector
                
                # Calculate new velocity and new position
                
                for p, cP in enumerate(self.cS):
    
                    cP.V = w[p] * cP.V + \
                                   c1 * R1[p, :] * (self.cB[p].X - cP.X) + \
                                   c2 * R2[p, :] * (self.cB[self.BI[p]].X - cP.X)
                    cP.X = cP.X + cP.V        
                    
                    # Correct position to the bounds
                    cP.X = np.clip(cP.X, self.lb, self.ub)       
                    
                # Get old fitness
                f_old = np.array([cP.f for cP in self.cS])
    
                # Evaluate swarm
                self.collective_evaluation(self.cS)
    
                for p, cP in enumerate(self.cS):
                    # Calculate dF
                    self.dF[p] = cP.f - f_old[p]
                    
                for p, cP in enumerate(self.cS):
                    # Update personal best
                    if cP <= self.cB[p]:
                        self.cB[p] = cP.copy()  
                
                # Update the overall best
                self.p_best = np.argmin(self.cB)
                
                # Update history
                #if np.min(self.cX).f < self.results.cHistory[-1].f:
                    #self.results.cHistory.append(np.min(self.cX).copy()) # superseded by progress_log()
                self._progress_log(remark='(PSO)')
                
                # Check stopping conditions
                if self._stopping_criteria():
                    break
                
                # Update swarm topology
                if np.max(self.dF) >= 0.0:
                    self.reinitialize_topology()
                    
                # Find best particles in neighbourhood 
                self.find_neighbourhood_best()
            
            """ FWA iteration """
            if 'FWA' in self.method:
                explosion_sparks = self.explosion()
                mutation_sparks = self.gaussian_mutation()
    
                #self.__mapping_rule(sparks, self.lb, self.ub, self.dimensions)
                for cS in (explosion_sparks + mutation_sparks):
                    
                    ilb = cS.X < self.lb
                    cS.X[ilb] = self.lb[ilb]
                    iub = cS.X > self.ub
                    cS.X[iub] = self.ub[iub]
    
                    self.cX = np.append(self.cX, [cS])
    
                self.collective_evaluation(self.cX)
    
                self.cX = np.sort(self.cX)[:n]
    
                # Update history
                #if np.min(self.cX).f < self.results.cHistory[-1].f:
                    #self.results.cHistory.append(np.min(self.cX).copy()) # superseded by progress_log()
                self._progress_log(remark='(FWA)')
    
                # Check stopping conditions
                if self._stopping_criteria():
                    break
            
            """ DE iteration """
            if 'DE' in self.method:
                K = 0 # memory index
                    
                S_CR = np.empty([0])
                S_F = np.empty([0])
                S_dF = np.empty([0])
                
                # find pbest
                top = max(round(np.size(self.Pop) * self.params['p_mutation']), 1)
                pbest = np.random.choice(np.sort(self.Pop)[0:top])
                
                for p, t in zip(self.Pop, self.Trials):
                    
                    # Update CR, F
                    r = np.random.randint(self.params['historical_memory_size'])
                    if np.isnan(self.M_CR[r]):
                        p.CR = 0
                    else:
                        p.CR = np.random.normal(self.M_CR[r], 0.1)
                        p.CR = np.clip(p.CR, 0, 1)
                    p.F = -1
                    while p.F <= 0:
                        p.F = min(cauchy.rvs(self.M_F[r], 0.1), 1)
                    
                    # Compute mutant vector
                    r1 = r2 = p
                    while r1 is r2 or r1 is p or r2 is p:
                        r1 = np.random.choice(self.Pop)
                        r2 = np.random.choice(np.append(self.Pop, self.A))
                    p.V = p.X + p.F * (pbest.X - p.X) + p.F * (r1.X - r2.X)
                    p.V = np.clip(p.V, (p.X + self.lb)/2, (p.X + self.ub)/2)
                    
                    # Compute trial vector
                    t.CR = p.CR
                    t.F = p.F
                    jrand = np.random.randint(self.dimensions)
                    for j in range(self.dimensions):
                        if np.random.rand() <= p.CR or j == jrand:
                            t.X[j] = p.V[j]
                        else:
                            t.X[j] = p.X[j]
    
                # Evaluate population
                self.collective_evaluation(self.Trials)
                
                # Survival for next generation
                for p, t in zip(self.Pop, self.Trials):
                    if t.f <= p.f:
                        p.X = np.copy(t.X)
                        p.f = t.f                    
                    # Update external archive
                    if t.f < p.f:
                        self.A = np.append(self.A, p)
                        if np.size(self.A) > round(np.size(self.Pop) * self.params['external_archive_size_factor']):
                            self.A = np.delete(self.A, 
                                               np.random.randint(np.size(self.A)))
                        S_CR = np.append(S_CR, t.CR) 
                        S_F = np.append(S_F, t.F)
                        S_dF = np.append(S_dF, p.f - t.f)
    
                # Memory update
                if np.size(S_CR) != 0 and np.size(S_F) != 0:
                    w = S_dF / np.sum(S_dF)
                    if np.isnan(self.M_CR[K]) or np.max(S_CR) < 1e-100:
                        self.M_CR[K] = np.nan
                    else:
                        self.M_CR[K] = np.sum(w * S_CR**2) / np.sum(w * S_CR)
                    self.M_F[K] = np.sum(w * S_F**2) / np.sum(w * S_F)
                    K += 1
                    if K >= self.params['historical_memory_size']:
                        K = 0
                        
                # Linear Population Size Reduction (LPSR)
                if self.method == 'LSHADE':
                    N_init = self.params['initial_population_size']
                    if self.maximum_evaluations:
                        N_new = round(((4 - N_init) / self.maximum_evaluations) \
                                            * self.evaluations_count + N_init)
                    else:
                        N_new = round(((4 - N_init) / self.iterations) \
                                            * self.it + N_init)
                    if N_new < np.size(self.Pop):
                        self.Pop = np.sort(self.Pop)[:N_new]
                        self.Trials = self.Trials[:N_new]          
                    
                # Update history
                # self.results.cHistory.append(self.best.copy()) # superseded by progress_log()
                self._progress_log(remark='(DE)')
                
                # Check stopping conditions
                if self._stopping_criteria():
                    break
            
            """ Methods cooperating """
            bests = {}
            if 'PSO' in self.method:
                bests['PSO'] = self.cB[self.p_best].f
            if 'FWA' in self.method:
                bests['FWA'] = np.min(self.cX).f
            if 'DE' in self.method:
                bests['DE'] = np.min(self.Pop).f 
            
            if min(bests, key=bests.get) == 'PSO': # PSO best is THE BEST
                if 'FWA' in self.method:
                    FWAworst = np.max(self.cX)
                    FWAworst.X = np.copy(self.best.X) 
                    FWAworst.f = self.best.f
                if 'DE' in self.method:
                    DEworst = np.max(self.Pop)
                    DEworst.X = np.copy(self.best.X) 
                    DEworst.f = self.best.f
                #print('PSO updated others')
            elif min(bests, key=bests.get) == 'FWA': # FWA best is THE BEST
                if 'PSO' in self.method:
                    PSOworst = np.max(self.cB)
                    PSOworst.X = np.copy(self.best.X) 
                    PSOworst.f = self.best.f 
                    PSOworst.dF = 0.0
                    self.find_neighbourhood_best()
                if 'DE' in self.method:
                    DEworst = np.max(self.Pop)
                    DEworst.X = np.copy(self.best.X) 
                    DEworst.f = self.best.f
                #print('FWA updated others')
            elif min(bests, key=bests.get) == 'DE': # FWA best is THE BEST
                if 'PSO' in self.method:
                    PSOworst = np.max(self.cB)
                    PSOworst.X = np.copy(self.best.X) 
                    PSOworst.f = self.best.f 
                    PSOworst.dF = 0.0
                    self.find_neighbourhood_best()
                if 'FWA' in self.method:
                    FWAworst = np.max(self.cX)
                    FWAworst.X = np.copy(self.best.X) 
                    FWAworst.f = self.best.f
                #print('DE updated others')
       
        return self.best
    


# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import time
import multiprocessing
import string


class Optimizer:
    """Base class for all optimization methods.

    Parameters
    ----------
    dimensions: int
        Number of dimensions of the search space i.e. number of optimization variables.
    evaluation_function: callable
        A function (or a callable class) which takes a Numpy:ndarray as argument and returns a Numpy:ndarray containing
        objectives' and constraint's values. Optionally, it can take a keyword argument :code:`s` in order to handle
        unique string generated by the optimizer.
    monitoring: str or None
        Adjust the monitoring style and level of output produced during the optimization. :code:`monitoring='None'` or
        :code:`monitoring=None` suppress all output. For displaying a per iteration output to the terminal use
        :code:`monitoring='basic'` which can display a colored output if supported by your terminal. For displaying a
         progress bar and dynamically updated best candidate state use :code:`monitoring='dashboard'`.
    best: CandidateState
        The best candidate state encountered in the optimization process.
    """

    def __init__(self):
        """Initialization for Optimizer class defines member variables and set initial values.

        Returns
        -------
        Optimizer
            Optimizer instance
        """

        self.dimensions = None
        self.evaluation_function = None
        self.number_of_processes = 1
        
        self.objectives = 1
        self.objective_weights = None
        self.objective_labels = None
        self.constraints = 0
        self.constraint_labels = None
        
        self.iterations = None
        self.evaluations_count = 0
        self.maximum_evaluations = None
        self.target_fitness = None
        
        self.lb = None
        self.ub = None
        self.best = None
        
        self.monitoring = None
        self._progress_log = None
        self._time1 = time.time()

        self.forward_unique_str = False
        self._unique_str_list = []
        self.it = 0
        
    def _init_optimizer(self):
        """A private method for Optimizer initialization which checks the types and values of Optimizer member
        parameters. Automatically sets optional values if they are not provided.

        Returns
        -------
        None
            Nothing
        """
        
        assert isinstance(self.dimensions, int) and self.dimensions > 0, \
            "optimizer.dimensions should be positive integer"
            
        if np.size(self.lb) == 1 and np.size(self.ub) == 1:
            self.lb = np.full(self.dimensions, self.lb)
            self.ub = np.full(self.dimensions, self.ub)
            
        assert np.size(self.lb) == self.dimensions and \
            np.size(self.ub) == self.dimensions, \
            "optimizer.lb and optimizer.ub should be of size optimizer.dimensions"

        assert callable(self.evaluation_function), "optimizer.evaluation_function needs to be callable"

        assert isinstance(self.objectives, int) and self.objectives > 0, \
            "optimizer.objectives should be positive integer"
        
        assert isinstance(self.constraints, int) and self.constraints >= 0, \
            "optimizer.constraints should be non-negative integer"
        
        if self.objective_weights is None:
            self.objective_weights = np.ones(self.objectives) / self.objectives   
        else:
            assert len(self.objective_weights) == self.objectives, \
                "optimizer.objective_weights list should contain number of elements equal to optimizer.objectives"
        
        if self.objective_labels is None:
            self.objective_labels = [f'obj_{o}' for o in range(self.objectives)]
        else:
            assert len(self.objective_labels) == self.objectives, \
                "optimizer.objective_labels list should contain number of strings equal to optimizer.objectives"
            
        if self.constraint_labels is None:
            self.constraint_labels = [f'cnstr_{c}' for c in range(self.constraints)]
        else:
            assert len(self.constraint_labels) == self.constraints, \
                "optimizer.constraint_labels list should contain number of strings equal to optimizer.constraints"

        assert (isinstance(self.number_of_processes, int) and self.number_of_processes > 0)\
               or self.number_of_processes == 'maximum', \
               "optimizer.number_of_processes should be positive integer or \'maximum\'"
        if self.number_of_processes == 'maximum':
            self.number_of_processes = multiprocessing.cpu_count()
        if self.number_of_processes > 1: 
            if self.monitoring == 'basic':
                self.log(f'Preparing a multiprocess pool for {self.number_of_processes} processes.')
            self.pool = multiprocessing.Pool(self.number_of_processes)

        if self.maximum_evaluations:
            assert isinstance(self.maximum_evaluations, int) and self.maximum_evaluations > 0, \
                "optimizer.maximum_evaluations should be positive integer"

        if self.iterations is None:
            self.iterations = 100 * self.dimensions

        assert isinstance(self.monitoring, str) or self.monitoring is None, \
            'optimizer.monitoring should be a string or None'

        if isinstance(self.monitoring, str):
            # Convert to lowercase string
            self.monitoring = self.monitoring.lower()
        if self.monitoring is None:
            self.monitoring = 'none'

        if self.monitoring == 'none':
            self._progress_log = lambda remark=None: None

        elif self.monitoring == 'basic':
            # self.progress_log = self._progress_log2
            from rich.console import Console
            self.rich_console = Console(highlight=False)
            self._progress_log = self.__progress_log_basic

        elif self.monitoring == 'dashboard':
            from rich.progress import Progress, BarColumn, TimeRemainingColumn, TimeElapsedColumn
            from rich.live import Live
            from rich.console import RenderGroup, Text
            # from rich.panel import Panel
            # from datetime import timedelta
            import indago

            self._progress_log = lambda remark=None: None
            self.__progress_bar = Progress(
                "[progress.description]{task.description}",
                # "{task.completed:} of {task.total}",
                BarColumn(),
                "[progress.percentage]{task.percentage:>3.0f}%",
                # "Elapsed: {task.time_remaining}",
                "Elapsed:",
                TimeElapsedColumn(),
                "Remaining:",
                TimeRemainingColumn(),
                # expand=True,
            )

            self.__progress_bar_tasks = [self.__progress_bar.add_task("Max. iterations",
                                                                      total=self.iterations)]
            if self.maximum_evaluations:
                self.__progress_bar_tasks.append(self.__progress_bar.add_task("Max. evaluations",
                                                                              total=self.maximum_evaluations))
            if self.target_fitness:
                self.__progress_bar_tasks.append(self.__progress_bar.add_task("Target fitness", total=1))

            def update_progress_bar():

                progress = [self.it]

                if self.maximum_evaluations:
                    progress.append(self.evaluations_count)

                if self.target_fitness and self.best:
                    f0 = None
                    for h in self.results.cHistory:
                        if np.all(h[2].C <= 0):
                            f0 = h[2].f
                            break

                    if f0:
                        if self.best.f > self.target_fitness:
                            p = (self.best.f - self.target_fitness) / (f0 - self.target_fitness)
                        else:
                            p = 0
                        p = 1 - np.log10(1 + 9 * p)
                    else:
                        p = 0
                    progress.append(p)

                for task, prog in zip(self.__progress_bar_tasks, progress):
                    self.__progress_bar.update(task, completed=prog)

                best = 'None'
                if self.best is not None:
                    best = str(self.best.X)
                    all_labels = self.objective_labels + self.constraint_labels
                    label_len = np.max([len(lbl) for lbl in all_labels]) + 3
                    for o, o_label in zip(self.best.O, self.objective_labels):
                        best += '\n[magenta]' + f'{o_label}: '.rjust(label_len) + f'{o:12.5e} [/magenta]'
                    for c, c_label in zip(self.best.C, self.constraint_labels):
                        if c <= 0:
                            best += '\n[green]' + f'{c_label}: '.rjust(label_len) + f'{c:12.5e}   => (satisfied)[/green]'
                        else:
                            best += '\n[red]' + f'{c_label}: '.rjust(label_len) + f'{c:12.5e}   => (violated)[/red]'
                    best += '\n' + 'Fitness: '.rjust(label_len) + f'{self.best.f:12.5e}'

                return RenderGroup(f'[b]INDAGO[/b] {indago.__version__}',
                                   f'{self.__class__.__name__} optimization running\n',
                                   self.__progress_bar,
                                   f'\nCurrent iteration:{self.it}\nTotal evaluations:{self.evaluations_count}\nBest:\n{best}')
            self.__update_progress_bar = update_progress_bar
            self.live = Live(self.__update_progress_bar())

        else:
            assert False, f'Unknown value for optimizer.monitoring parameter ({self.monitoring})'
        self.evaluations_count = 0
        self.it = 0

        # Initialize history
        self.results = OptimizationResults(self)
        self.results.cHistory = []
        self.best = None

    def _check_params(self, mandatory_params, optional_params, defined_params):
        """Private method which checks if optimizer parameters are defined in Optimizer.params dictionary.
        Should be called in initializations of derived Optimizer classes. It asserts if any of mandatory parameters is
        missing and prints a warning if unknown/excessive parameter is provided.

        Parameters
        ----------
        mandatory_params : list of str
            A list of mandatory parameter names/keys for given method.
        optional_params: list of str
            A list of optional parameter names/keys for given method.
        defined_params: list of str
            A list of method's parameters names/keys defined in Optimizer.params dictionary.

        Returns
        -------
        None
            Nothing
        """
        for param in mandatory_params:
            assert param in defined_params, f'Error: Missing parameter {param}'

        for param in defined_params:
            if param not in mandatory_params and param not in optional_params:
                print(f'Warning: Excessive parameter {param}')

    @staticmethod
    def title(msg, line_length=60):
        """Utility function for consistently displaying titles in the output. The text of the title is decorated and
        printed on the center of the screen.

        Parameters
        ----------
        msg: str
            A text of the title.
        line_length: int
            A length (number of characters) of the line used for drawing the decoration border and
        centering the title. Default is 60.

        Returns
        -------
        None
            Nothing
        """
        t = '-' * line_length
        t += '\n' + msg.center(line_length) + '\n'
        t += '-' * line_length
        print(t)

    @staticmethod
    def log(msg, indent=0):
        """Utility function for consistently displaying messages in the output.

        Parameters
        ----------
        msg: str
            Message to be displayed in the output.
        indent: int
            Number of characters used for indentation of the message. Default is 0.
        Returns
        -------
        None
            Nothing
        """
        print(' ' * indent + msg)

    def __progress_log_basic(self, remark=None):
        line = f'iter: {self.it:{int(np.ceil(np.log10(self.iterations))) + 1}d}/{self.iterations}'
        line += f', eval: {self.evaluations_count:10d}'
        if self.maximum_evaluations:
            line += f'/{self.maximum_evaluations}'
        for io, ol in enumerate(self.objective_labels):
            line += f', {ol}: [blue]{self.best.O[io]}[/blue]'
        for cl, cv in zip(self.constraint_labels, self.best.C):
            if cv > 0:
                line += f', {cl}: [red]{cv}[/red]'
            else:
                line += f', {cl}: [green]{cv}[/green]'
        line += f', fit: [magenta bold underline]{self.best.f}[/magenta bold underline]'
        if remark:
            line += f' {remark}'
        self.rich_console.print(line)
        #self.log(line)

    def tic(self):
        """Utility function for measuring the computational time. Optimizer.tic() is used for starting the stopwatch.
        It is used in pair with Optimizer.toc() method which stops the stopwatch.

        Returns
        -------
        Float
            Return the recorded current time in seconds since the Epoch.
        """
        self._time1 = time.time()
        return self._time1

    def toc(self, msg='', silent=False):
        """Utility function for measuring the computational time. Optimizer.toc() stops the stopwatch and returns
        time elapsed since Optimizer.tic() was triggered.

        Parameters
        ----------
        msg: str, default=''
            A message which is displayed with the measured time. Not used if silent=True.
        silent: bool, default=True
            An option for disabling the output. Default is False. If True, the value of elapsed time is returned but no
            message is shown.

        Returns
        -------
        elapsed_time: float
            Time elapsed since calling the last Optimizer.toc() in seconds.
        """

        s = time.time() - self._time1
        
        if not silent:
            if msg == '':
                msg = 'Elapsed time:'

            if s < 1:
                dur = f'{s*1e3:.3f} milliseconds.'
            else:
                dur = f'{s:.3f} seconds.'
            self.log(f'{msg} {dur}')
            
        return s

    def _gen_unique_str(self):
        """A private method for generating a unique string of 16 characters used when forwarding a unique string to the
        evaluation function is enabled (Optimizer.forward_unique_str = True). The generated string is guaranteed to be
        unique within an optimizer.

        Returns
        -------
        s: str
            A unique string.
        """
        while True:
            _s = ''.join(np.random.choice([c for c in string.ascii_lowercase]) for _ in range(16))
            if _s not in self._unique_str_list:
                self._unique_str_list.append(_s)
                return _s

    def optimize(self, seed=None):
        """A method which starts the optimization. The method wraps :code:`_run` method of the optimizer's subclass.

        Parameters
        ----------
        seed: int or None
            A random seed. Use the same value for reproducing identical stochastic procedures.

        Returns
        -------
        optimum: OptimizationResults
            The best solution found in the optimization.
        """
        self._init_optimizer()
        np.random.seed(seed)

        if self.monitoring == 'dashboard':
            with self.live:
                r = self._run()
        else:
            r = self._run()

        return r

    def _stopping_criteria(self):
        """A private method used to evaluate all specified stopping conditions.

        Returns
        -------
        stop: bool
            A :code:`True` or :code:`False`, whether the optimization should stop.
        """

        # Stop if fitness threshold achieved
        if self.target_fitness:
            if self.best.f < self.target_fitness and np.all(self.best.C <= 0):
                return True

        # Stop if maximum number of evaluations surpassed
        if self.maximum_evaluations:
            if self.evaluations_count >= self.maximum_evaluations:
                return True

        return False

    def __getstate__(self):
        self_dict = self.__dict__.copy()
        del self_dict['pool']
        return self_dict

    def __setstate__(self, state):
        self.__dict__.update(state)

    def _multiprocess_evaluate(self, candidates):
        """A private method used for calling parallel evaluation of multiple candidates. It relies on multiprocessing
        pool using map and starmap methods. The objectives, constraints and fitness of candidates in the list is updated
        after the evaluation.

        Parameters
        ----------
        candidates: list of CandidateState
            A list of candidates to be evaluated.

        Returns
        -------
        candidates: list of CandidateState
            A list of evaluated candidates.
        """

        if self.forward_unique_str:
            result = self.pool.starmap(self.evaluation_function,
                                       [tuple([c.X, self._gen_unique_str()]) for c in candidates])
        else:
            result = self.pool.map(self.evaluation_function, [c.X for c in candidates])

        if self.objectives == 1 and self.constraints == 0:
            # Fast evaluation
            for p in range(len(candidates)):
                candidates[p].f = result[p]
                candidates[p].O[0] = result[p]
        else:
            # Full evaluation
            for p in range(len(candidates)):
                for io in range(self.objectives):
                    candidates[p].O[io] = result[p][io]
                candidates[p].f = np.dot(candidates[p].O, self.objective_weights)
                for ic in range(self.constraints):
                    candidates[p].C[ic] = result[p][self.objectives + ic]

        return candidates  # TODO do we need to return candidates or they can be updated in-place?

    def collective_evaluation(self, candidates):
        """Function used for evaluation of multiple candidates which automatically conducts parallel or serial
        evaluation and forwards a unique string to the evaluation function. Evaluation is performed in-place and the
        candidates provided via argument are updated.

        Parameters
        ----------
        candidates: list of CandidateState
            A list of candidates to be evaluated.

        Returns
        -------
        None
            Nothing
        """

        n = len(candidates)
        self.evaluations_count += n

        if self.number_of_processes > 1:
            self._multiprocess_evaluate(candidates)
        else:
            if self.forward_unique_str:
                for p in range(n):
                    candidates[p].evaluate(self._gen_unique_str())
            else:
                for p in range(n):
                    candidates[p].evaluate()

        # Determine new best candidate state
        if self.best is None:
            self.best = np.min(candidates).copy()
            self.results.cHistory.append((self.it, self.evaluations_count, self.best))
        else:
            # self.best = np.min(np.append(candidates, [self.best])).copy()
            candidates_best = np.min(candidates)
            if candidates_best <= self.best:
                self.best = candidates_best.copy()
                self.results.cHistory.append((self.it, self.evaluations_count, self.best))

        if self.monitoring == 'dashboard':
            self.live.update(self.__update_progress_bar())

    def multiprocess_evaluate_test(self, number_of_candidates, seed=None):
        # Function for testing purpose only - remove it and use it in the test sources!

        self._init_optimizer()
        if None:
            np.random.seed() # Possible fix for some seed issues in parallel execution
        else:
            np.random.seed(seed)

        candidates = np.array([CandidateState(self) for _ in range(number_of_candidates)], dtype=CandidateState)

        for p in range(number_of_candidates):
            candidates[p].X = np.random.uniform(self.lb, self.ub)

        self.collective_evaluation(candidates)

        return candidates


class CandidateState:
    """Candidate solution for optimization problem"""
    
    def __init__(self, optimizer: Optimizer):
        self._optimizer = optimizer
        self.X = np.zeros(optimizer.dimensions)
        self.O = np.zeros(optimizer.objectives) * np.nan
        self.C = np.zeros(optimizer.constraints) * np.nan
        self.f = np.nan
                    
        # Evaluation
        if self._optimizer.objectives == 1 and self._optimizer.constraints == 0:
            self._evaluate = self._eval_fast
        else:
            self._evaluate = self._eval_full
        
        # Comparison operators
        if self._optimizer.objectives == 1 and self._optimizer.constraints == 0:
            self._eq_fn = CandidateState._eq_fast
            self._lt_fn = CandidateState._lt_fast 
            #self.__gt__ = self._gt_fast
        else:
            self._eq_fn = CandidateState._eq_full
            self._lt_fn = CandidateState._lt_full 
            #self.__gt__ = self._gt_full
        
        
    def copy(self):
        
        cP = CandidateState(self._optimizer)        
        # Real copy for ndarrays
        cP.X = np.copy(self.X)
        cP.O = np.copy(self.O)
        cP.C = np.copy(self.C)
        cP.f = self.f
        return cP
    
    def __str__(self): 
        nch = np.max([np.size(self.X), np.size(self.O), np.size(self.C)]) 
        nch = np.min([8, nch])
        #print(nch)
        s = '-' * (12 * nch + nch + 20)
        
        for i in range(int(np.ceil(np.size(self.X)/nch))):
            if i == 0:
                s += '\n' + 'X: '.rjust(20)
            else:
                s += '\n' + ' ' * 20
            s += ' '.join([f'{x:12.5e}' for x in self.X[i*nch:(i+1)*nch]])

        all_labels = self._optimizer.objective_labels + self._optimizer.constraint_labels
        label_len = np.max([len(lbl) for lbl in all_labels]) + 3
        for o, o_label in zip(self.O, self._optimizer.objective_labels):
            s += '\n' + f'{o_label}: '.rjust(label_len) + f'{o:12.5e}'
        for c, c_label in zip(self.C, self._optimizer.constraint_labels):
            s += '\n' + f'{c_label}: '.rjust(label_len) + f'{c:12.5e}'
        s += '\n' + 'Fitness: '.rjust(label_len) + f'{self.f:12.5e}' + \
             '\n' + '-' * (12 * nch + nch + 20)
        return s
    
    # Equality operator
    def __eq__(self, other): 
        return self._eq_fn(self, other) 
    
    @staticmethod
    def _eq_fast(a, b): 
        return a.f == b.f

    @staticmethod
    def _eq_full(a, b):
        return np.sum(np.abs(a.X - b.X)) + np.sum(np.abs(a.O - b.O)) + np.sum(np.abs(a.C - b.C)) == 0.0

    # Inequality operator
    def __ne__(self, other):
        return self.f != other.f
    
    # Less-then operator
    def __lt__(self, other):
        return self._lt_fn(self, other)    

    def evaluate(self, s=None):
        return self._evaluate(s)

    @staticmethod
    def _lt_fast(a, b): 
        #print('_lt_fast')
        return a.f < b.f   
    
    @staticmethod     
    def _lt_full(a, b):  
        #print("_lt_full")               
        if np.sum(a.C > 0) == 0 and np.sum(b.C > 0) == 0: 
            # Both are feasible
            # Better candidate is the one with smaller fitness
            return a.f < b.f
        
        elif np.sum(a.C > 0) == np.sum(b.C > 0): 
            # Both are unfeasible and break same number of constraints
            # Better candidate is the one with smaller sum of constraint values
            return np.sum(a.C) < np.sum(b.C)
        
        else:
            # The number of unsatisfied constraints is not the same
            # Better candidate is the one which breaks fewer constraints
            return np.sum(a.C > 0) < np.sum(b.C > 0)
       
    
    def __gt__(self, other):
        #print('__gt__')
        #return self.f > other.f
        return not (self._lt_fn(self, other) or self._eq_fn(self, other))
    """
    def _gt_fast(self, other):
        return self.f > other.f
    def _gt_full(self, other): 
        return not (self.__eq__(other) or self.__lt__(other))
    """    
    
    def __le__(self, other):
        return self._lt_fn(self, other) or self.__eq__(other)
    
    def __ge__(self, other):
        return self.__gt__(other) or self.__eq__(other)

    def set_X(self, X: np.ndarray):

        assert np.size(self.X) == np.size(X), 'Unexpected size of optimization vector X'
        self.X = X
        
    def _eval_full(self, s=None):

        if s is None:
            oc = self._optimizer.evaluation_function(self.X)
        else:
            oc = self._optimizer.evaluation_function(self.X, s)

        for io in range(self._optimizer.objectives):
            self.O[io] = oc[io]
        
        self.f = np.dot(self.O, self._optimizer.objective_weights)
        
        for ic in range(self._optimizer.constraints):
            self.C[ic] = oc[self._optimizer.objectives + ic]

        return self.O, self.C, self.f
        
    def _eval_fast(self, s=None):

        if s is None:
            f = self._optimizer.evaluation_function(self.X)
        else:
            f = self._optimizer.evaluation_function(self.X, s)

        self.f = f
        self.O[0] = f
        return self.O, self.C, self.f


class OptimizationResults:
    """Data holder for optimization results."""
    
    def __init__(self, optimizer: Optimizer):
        
        self.optimizer = optimizer
        self.cHistory = []
        
    def plot_convergence(self, axes=None):
        """Plot convergence

        Parameters
        ----------
        axes :

        Returns
        -------

        """
        
        if axes is None:
            fig = plt.figure(constrained_layout=True)
            spec = gridspec.GridSpec(ncols=1, nrows=2, figure=fig)
            axo = fig.add_subplot(spec[0])
            axc = fig.add_subplot(spec[1], sharex=axo)
        else:
            axo, axc = axes
        
        for io in range(self.optimizer.objectives):
            O = np.array([[ev, cB.O[io]] for (it, ev, cB) in self.cHistory])
            axo.plot(O[:, 0], O[:, 1], label=self.optimizer.objective_labels[io], lw=1)
            
        F = np.array([[ev, cB.f] for (it, ev, cB) in self.cHistory])
        axo.plot(F[:, 0], F[:, 1], label='f', lw=2, ls='-', c='b')
        axo.legend()
        
        for ic in range(self.optimizer.constraints):
            C = np.array([[ev, cB.C[ic]] for (it, ev, cB) in self.cHistory])
            axc.plot(C[:, 0], C[:, 1], label=self.optimizer.constraint_labels[ic], lw=1)

            IX = C[:, 1] > 0
            C = C[IX, :]

            # nc = np.sum(C>0)
            # axc.plot(I[:nc], C[:nc], c='r', ls='-', lw=2)
            # if nc == 1:
            #     axc.plot(I[:nc], C[:nc],'ro')
                            
        axc.legend()       
        axo.set_ylabel('Objectives')
        axc.set_xlabel('Evaluations')
        axc.set_ylabel('Constraints')
        #ax.set_yscale('log')
        


# -*- coding: utf-8 -*-

import numpy as np
from .optimizer import Optimizer, CandidateState 
from scipy.stats import cauchy

"""
R. Tanabe and A. S. Fukunaga, “Improving the search performance of SHADE using 
linear population size reduction”, in Proceedings of the 2014 IEEE Congress 
on EvolutionaryComputation (CEC), pp. 1658–1665, Beijing, China, July 2014.
"""


class Solution(CandidateState):
    """DE solution class"""
    
    def __init__(self, optimizer: Optimizer):
        CandidateState.__init__(self, optimizer)
        #super(Particle, self).__init__(optimizer) # ugly version of the above
        
        self.CR = None
        self.F = None
        self.V = np.zeros([optimizer.dimensions]) * np.nan # mutant vector


class DE(Optimizer):
    """Differential Evolution class"""

    def __init__(self):
        """Initialization"""
        Optimizer.__init__(self)
        #super(PSO, self).__init__() # ugly version of the above

        self.X0 = None
        self.method = 'SHADE'
        self.params = {}

    def _check_params(self):
        defined_params = list(self.params.keys())
        mandatory_params, optional_params = [], []

        if self.method == 'SHADE':
            mandatory_params = 'initial_population_size external_archive_size_factor historical_memory_size p_mutation'.split()
            if 'initial_population_size' not in self.params:
                self.params['initial_population_size'] = self.dimensions * 18
                defined_params += 'initial_population_size'.split()
            if 'external_archive_size_factor' not in self.params:
                self.params['external_archive_size_factor'] = 2.6
                defined_params += 'external_archive_size_factor'.split()
            if 'historical_memory_size' not in self.params: # a.k.a. H
                self.params['historical_memory_size'] = 6
                defined_params += 'historical_memory_size'.split()
            if 'p_mutation' not in self.params:
                self.params['p_mutation'] = 0.11
                defined_params += 'p_mutation'.split()    
            optional_params = ''.split()
        elif self.method == 'LSHADE':
            mandatory_params = 'initial_population_size external_archive_size_factor historical_memory_size p_mutation'.split()
            if 'initial_population_size' not in self.params:
                self.params['initial_population_size'] = self.dimensions * 18
                defined_params += 'initial_population_size'.split()
            if 'external_archive_size_factor' not in self.params:
                self.params['external_archive_size_factor'] = 2.6
                defined_params += 'external_archive_size_factor'.split()
            if 'historical_memory_size' not in self.params: # a.k.a. H
                self.params['historical_memory_size'] = 6
                defined_params += 'historical_memory_size'.split()
            if 'p_mutation' not in self.params:
                self.params['p_mutation'] = 0.11
                defined_params += 'p_mutation'.split()  
            optional_params = ''.split()
        else:
            assert False, f'Unknown method! {self.method}'

        Optimizer._check_params(self, mandatory_params, optional_params, defined_params)
        
        assert isinstance(self.params['initial_population_size'], int) \
            and self.params['initial_population_size'] > 0, \
            "initial_population_size should be positive integer"
        assert self.params['external_archive_size_factor'] > 0, \
            "external_archive_size should be positive"
        assert isinstance(self.params['historical_memory_size'], int) \
            and self.params['historical_memory_size'] > 0, \
            "historical_memory_size should be positive integer"
        
    def _init_method(self):
        
        # Bounds for position
        self.lb = np.array(self.lb)
        self.ub = np.array(self.ub)

        # Generate a population
        self.Pop = np.array([Solution(self) for c in \
                             range(self.params['initial_population_size'])], dtype=Solution)
        
        # Generate a trial population
        self.Trials = np.array([Solution(self) for c in \
                                range(self.params['initial_population_size'])], dtype=Solution)
        
        # Initalize Archive
        self.A = np.empty([0])
        
        # Prepare historical memory
        self.M_CR = np.full(self.params['historical_memory_size'], 0.5)
        self.M_F = np.full(self.params['historical_memory_size'], 0.5)

        # Generate initial positions
        for i in range(self.params['initial_population_size']):
            
            # Random position
            self.Pop[i].X = np.random.uniform(self.lb, self.ub)
            
            # Using specified particles initial positions
            if self.X0 is not None:
                if i < np.shape(self.X0)[0]:
                    self.Pop[i].X = self.X0[i]

        # Evaluate
        self.collective_evaluation(self.Pop)
        
        self._progress_log()
        
    def _run(self):
        self._check_params()
        self._init_method()
        
        k = 0 # memory index

        for self.it in range(1, self.iterations + 1):
            
            S_CR = np.empty([0])
            S_F = np.empty([0])
            S_dF = np.empty([0])
            
            # find pbest
            top = max(round(np.size(self.Pop) * self.params['p_mutation']), 1)
            pbest = np.random.choice(np.sort(self.Pop)[0:top])
            
            for p, t in zip(self.Pop, self.Trials):
                
                # Update CR, F
                r = np.random.randint(self.params['historical_memory_size'])
                if np.isnan(self.M_CR[r]):
                    p.CR = 0
                else:
                    p.CR = np.random.normal(self.M_CR[r], 0.1)
                    p.CR = np.clip(p.CR, 0, 1)
                p.F = -1
                while p.F <= 0:
                    p.F = min(cauchy.rvs(self.M_F[r], 0.1), 1)
                
                # Compute mutant vector
                r1 = r2 = p
                while r1 is r2 or r1 is p or r2 is p:
                    r1 = np.random.choice(self.Pop)
                    r2 = np.random.choice(np.append(self.Pop, self.A))
                p.V = p.X + p.F * (pbest.X - p.X) + p.F * (r1.X - r2.X)
                p.V = np.clip(p.V, (p.X + self.lb)/2, (p.X + self.ub)/2)
                
                # Compute trial vector
                t.CR = p.CR
                t.F = p.F
                jrand = np.random.randint(self.dimensions)
                for j in range(self.dimensions):
                    if np.random.rand() <= p.CR or j == jrand:
                        t.X[j] = p.V[j]
                    else:
                        t.X[j] = p.X[j]

            # Evaluate population
            self.collective_evaluation(self.Trials)
            
            # Survival for next generation
            for p, t in zip(self.Pop, self.Trials):
                if t.f <= p.f:
                    p.X = np.copy(t.X)
                    p.f = t.f                    
                # Update external archive
                if t.f < p.f:
                    self.A = np.append(self.A, p)
                    if np.size(self.A) > round(np.size(self.Pop) * self.params['external_archive_size_factor']):
                        self.A = np.delete(self.A, 
                                           np.random.randint(np.size(self.A)))
                    S_CR = np.append(S_CR, t.CR) 
                    S_F = np.append(S_F, t.F)
                    S_dF = np.append(S_dF, p.f - t.f)

            # Memory update
            if np.size(S_CR) != 0 and np.size(S_F) != 0:
                w = S_dF / np.sum(S_dF)
                if np.isnan(self.M_CR[k]) or np.max(S_CR) < 1e-100:
                    self.M_CR[k] = np.nan
                else:
                    self.M_CR[k] = np.sum(w * S_CR**2) / np.sum(w * S_CR)
                self.M_F[k] = np.sum(w * S_F**2) / np.sum(w * S_F)
                k += 1
                if k >= self.params['historical_memory_size']:
                    k = 0
                    
            # Linear Population Size Reduction (LPSR)
            if self.method == 'LSHADE':
                N_init = self.params['initial_population_size']
                if self.maximum_evaluations:
                    N_new = round(((4 - N_init) / self.maximum_evaluations) \
                                        * self.evaluations_count + N_init)
                else:
                    N_new = round(((4 - N_init) / self.iterations) \
                                        * self.it + N_init)
                if N_new < np.size(self.Pop):
                    self.Pop = np.sort(self.Pop)[:N_new]
                    self.Trials = self.Trials[:N_new]          
                
            # Update history
            # self.results.cHistory.append(self.best.copy()) # superseded by progress_log()
            self._progress_log()
            
            # Check stopping conditions
            if self._stopping_criteria():
                break

        return self.best
